import csv
import json
import os
from glob import glob
from pathlib import Path

import tomli
import yaml
from traitlets.config.application import get_config

from pushcart.release import Release

user = Release.instance().user
repo = Release.instance().repo
log = Release.instance().log

loaders = {".json": json.load, ".toml": tomli.load, ".yaml": yaml.load}

# TODO: See if the /Workspace/Repos part cannot be inferred or parameterized

all_pipelines_dir = f"/Workspace/Repos/{user}/{repo}/pipelines"
log.debug(f"Pulling pipeline configuration files from {all_pipelines_dir}")


def __enrich_sources_config(pipeline_name: str, config_sources: list) -> list:
    sources = []

    for s in config_sources:
        source_dict = s

        if isinstance(s.get("params"), dict):
            source_dict["params"] = json.dumps(s["params"])
        sources.append(source_dict)

    return sources


def __enrich_transformations_config(
    pipeline_name: str, config_transformations: list
) -> list:
    transformations = []

    for t in config_transformations:
        if t.get("config"):
            with open(
                os.path.join(all_pipelines_dir, pipeline_name, t["config"]), "r"
            ) as csv_file:
                csv_reader = csv.DictReader(csv_file)

                for row in csv_reader:
                    for k, v in row.items():
                        if k == "column_order":
                            row[k] = int(v) if len(v) > 0 else None
                        elif k == "not_null":
                            row[k] = bool(v) if len(v) > 0 else None

                    row["from"] = t["from"]
                    row["into"] = t["into"]

                    transformations.append(row)

        else:
            transformations.append(t)

    return transformations


def __enrich_destinations_config(pipeline_name: str, config_destinations: list) -> list:
    return config_destinations


def __enrich_pipeline_config(pipeline_name: str, config_dict: dict) -> dict:
    enriched_config_dict = {}

    enrichment_func = {
        "sources": __enrich_sources_config,
        "transformations": __enrich_transformations_config,
        "destinations": __enrich_destinations_config,
    }

    for stage_name in ["sources", "transformations", "destinations"]:
        stage = config_dict.get(stage_name, [])

        stage = enrichment_func[stage_name](pipeline_name, stage)
        stage = [{k: None if v == "" else v for k, v in elem.items()} for elem in stage]

        for stage_element in stage:
            stage_element["pipeline_name"] = pipeline_name

        enriched_config_dict[stage_name] = stage

    return enriched_config_dict


def __load_config_file(config_file_path: str) -> dict:
    config_dict = {}

    with open(config_file_path, "r") as config_file:
        ext = Path(config_file_path).suffix
        config_dict = loaders[ext](config_file)

    return config_dict


def collect_cluster_configs(pipeline_names: list) -> list:
    log.info("Collected cluster configurations for pipelines:")

    cluster_configs_dir = os.path.join(Path(all_pipelines_dir).parent, "clusters")

    cluster_configs = []
    for p in pipeline_names:
        all_cluster_config_files = set(glob(f"{cluster_configs_dir}/*"))

        search_for_files = [
            os.path.join(cluster_configs_dir, f"{p}{ext}") for ext in loaders.keys()
        ]

        found_cluster_configs = all_cluster_config_files.intersection(search_for_files)
        first_match = found_cluster_configs.pop()

        if found_cluster_configs:
            log.warning(
                f"Found more than one potential cluster configuration for {p}. Choosing {first_match}"
            )

        cluster_configs.append({p: __load_config_file(first_match)})
        log.info(f"- {p}")

    if not cluster_configs:
        log.info("- None")

    return cluster_configs


def collect_pipeline_configs(pipelines_dir: str) -> list:
    pipeline_configs = []
    pipeline_files = []

    json_files = glob(f"{pipelines_dir}/**/*.json")
    pipeline_files.extend(json_files)
    if json_files:
        log.info(f"Found {len(json_files)} JSON files")

    toml_files = glob(f"{pipelines_dir}/**/*.toml")
    pipeline_files.extend(toml_files)
    if toml_files:
        log.info(f"Found {len(toml_files)} TOML files")

    yaml_files = glob(f"{pipelines_dir}/**/*.yaml")
    pipeline_files.extend(yaml_files)
    if yaml_files:
        log.info(f"Found {len(yaml_files)} YAML files")

    pipeline_names = set()
    for config_file_path in pipeline_files:
        pipeline_name = Path(config_file_path).parent.stem
        pipeline_names.add(pipeline_name)

        config_dict = __enrich_pipeline_config(
            pipeline_name, __load_config_file(config_file_path)
        )
        pipeline_configs.append(config_dict)

    cluster_configs = collect_cluster_configs(pipeline_names)

    stage_records = {
        "clusters": [],
        "sources": [],
        "transformations": [],
        "destinations": [],
    }

    stage_records["clusters"] = cluster_configs

    for config in pipeline_configs:
        for stage_name, stage in config.items():
            stage_records[stage_name].extend(stage)

    if stage_records["clusters"]:
        log.info(f"Loaded {len(stage_records['clusters'])} cluster configurations")
    if stage_records["sources"]:
        log.info(f"Loaded {len(stage_records['sources'])} sources")
    if stage_records["transformations"]:
        log.info(f"Loaded {len(stage_records['transformations'])} transformations")
    if stage_records["destinations"]:
        log.info(f"Loaded {len(stage_records['destinations'])} destinations")

    return stage_records


p = collect_pipeline_configs(all_pipelines_dir)
c = get_config()

for stage_name, stage in p.items():
    setattr(c.PipelineConfiguration, stage_name, stage)



from traitlets import Dict, HasTraits

from pushcart.configuration.validation.common import ExistingDeltaPath, ViewName


class DeltaSourceConfiguration(HasTraits):
    """
    Stores pipeline source configuration for a path containing a Delta table.
    """

    from_dict = Dict(
        default_value={},
        help="Dictionary containing the configuration for the pipeline",
        allow_none=True,
        per_key_traits={
            "from": ExistingDeltaPath(),
            "into": ViewName(),
        },
    )



from traitlets import Dict, HasTraits

from pushcart.configuration.validation.common import TableName, ViewName


class TableSource(HasTraits):
    config = Dict(per_key_traits={"from": TableName, "into": ViewName})

    def get_increment(self):
        pass



if __package__:
    from pushcart import dbutils, spark

import os
from logging import ERROR, INFO
from pathlib import Path

from databricks_cli.clusters.api import ClusterApi
from databricks_cli.configure.config import _get_api_client
from databricks_cli.configure.provider import DatabricksConfig
from databricks_cli.pipelines.api import PipelinesApi
from traitlets import Bool, Unicode
from traitlets.config import Application

from pushcart.configuration import PipelineConfiguration


class Release(Application):
    classes = [PipelineConfiguration]
    config_file = Unicode(
        default_value="release", help="release configuration file base name"
    ).tag(config=True)

    host = Unicode(allow_none=True, help="URL of Databricks instance")
    token = Unicode(allow_none=True, help="User token to log into Databricks")
    user = Unicode(allow_none=False, help="Databricks Repos user where repo is cloned")
    repo = Unicode(allow_none=False, help="Name of Git repository for DLT pipelines")
    non_destructive = Bool(
        default_value=False,
        allow_none=True,
        help="Remove or ignore obsolete DLT pipelines",
    )

    aliases = {
        "sources": "PipelineConfiguration.sources",
        "transformations": "PipelineConfiguration.transformations",
        "destinations": "PipelineConfiguration.destinations",
        ("c", "config-file"): "Release.config_file",
        "log-level": "Release.log_level",
        "host": "Release.host",
        "token": "Release.token",
        "user": "Release.user",
        "repo": "Release.repo",
        ("nd", "non-destructive"): "Release.non_destructive",
    }

    flags = {
        "info": ({"Release": {"log_level": INFO}}, "Set log level to INFO"),
        "error": ({"Release": {"log_level": ERROR}}, "Set log level to ERROR"),
        "true": (
            {"Release": {"non_destructive": True}},
            "Ignore obsolete DLT pipelines",
        ),
        "false": (
            {"Release": {"non_destructive": False}},
            "Remove obsolete DLT pipelines",
        ),
    }

    @classmethod
    def get_subapp_instance(cls, app: Application) -> Application:
        """
        Needs to run if the release is a subapp, or if in the presence of the
        IPython Kernel (i.e. from Databricks). Since Application is a singleton,
        we need to clear main app
        """
        app.clear_instance()
        return cls.instance(parent=app)

    def initialize(self, argv=None):
        super().initialize(argv=argv)

        self.log_format = "%(asctime)s [%(name)s] %(levelname)s: %(message)s"
        self.log_datefmt = "%Y-%m-%dT%H:%M:%S%z"

        if __package__:
            cwd = Path(__file__).parent
        else:
            notebook_path = (
                dbutils.notebook.entry_point.getDbutils()
                .notebook()
                .getContext()
                .notebookPath()
                .get()
            )
            cwd = os.path.join(
                "/Workspace",
                os.path.dirname(os.path.abspath(notebook_path)).lstrip("/"),
            )

        self.log.info(f"Running release from {cwd}")

        if self.repo:
            spark.conf.set("pushcart.repo_name", self.repo)

        if self.user:
            spark.conf.set("pushcart.repo_user", self.user)

        if self.config_file:
            self.load_config_file(
                self.config_file, [os.path.join(cwd, "configuration", "presets")]
            )

    def start(self):
        self.__create_backend_objects()

        client = self.__get_api_client(self.host, self.token)

        workspace_pipelines = self.get_workspace_pipelines_list()
        scheduled_pipelines = self.get_scheduled_pipelines_list(client)
        new_pipelines = self.get_new_pipelines_list(
            workspace_pipelines, scheduled_pipelines
        )
        self.create_new_pipelines(client, new_pipelines)

        obsolete_pipelines = self.get_obsolete_pipelines_list(
            workspace_pipelines, scheduled_pipelines
        )

        if self.non_destructive:
            if obsolete_pipelines:
                self.log.warning(
                    f"Release run in non-destructive mode. Obsolete DLT pipelines ignored:"
                )
                for p in obsolete_pipelines:
                    self.log.warning(f"- {p}")

            return

        self.delete_obsolete_pipelines(client, obsolete_pipelines)

    def __create_backend_objects(self):
        schemas = {
            "sources": "struct<pipeline_name:string,from:string,type:string,into:string,params:string>",
            "transformations": "struct<pipeline_name:string,from:string,into:string,column_order:int,source_column_name:string,source_column_type:string,dest_column_name:string,dest_column_type:string,transform_function:string,sql_query:string,default_value:string,not_null:boolean>",
            "destinations": "struct<pipeline_name:string,from:string,into:string,path:string,keys:array<string>,partition_cols:array<string>,sequence_by:string>",
        }

        spark.sql("CREATE DATABASE IF NOT EXISTS metadata")

        for stage_name, schema in schemas.items():
            stage_df = spark.createDataFrame(
                self.config.PipelineConfiguration[stage_name], schema=schema
            )
            stage_df.write.saveAsTable(
                f"metadata.{stage_name}", format="delta", mode="overwrite"
            )

            self.log.info(f"Wrote {stage_name} metadata table.")

    @staticmethod
    def __get_api_client(host: str, token: str) -> DatabricksConfig:
        config = DatabricksConfig.from_token(host, token, False)

        return _get_api_client(config)

    @staticmethod
    def __get_api_node_types(client: DatabricksConfig) -> list:
        ca = ClusterApi(client)

        return ca.list_node_types().get("node_types", [])

    @staticmethod
    def __get_api_pipeline_list(client: DatabricksConfig) -> list:
        pa = PipelinesApi(client)

        return pa.list()

    def __delete_pipeline(self, client: DatabricksConfig, pipeline_id: str) -> None:
        pa = PipelinesApi(client)
        pa.delete(pipeline_id)

        self.log.info(f"Deleted {pipeline_id}")

    def __create_pipeline(self, client: DatabricksConfig, settings: dict) -> str:
        pa = PipelinesApi(client)
        pipeline = pa.create(settings, os.path.curdir, allow_duplicate_names=False)

        self.log.info(
            f"Created pipeline: {settings['name']} ({pipeline['pipeline_id']})"
        )
        return pipeline["pipeline_id"]

    def get_workspace_pipelines_list(self) -> list:
        src_df = spark.table("metadata.sources").select("pipeline_name")
        transf_df = spark.table("metadata.transformations").select("pipeline_name")
        dest_df = spark.table("metadata.destinations").select("pipeline_name")

        pipelines_df = src_df.union(transf_df).union(dest_df).distinct()
        workspace_pipelines = pipelines_df.pandas_api().get("pipeline_name").tolist()

        self.log.debug("Found workspace pipelines:")
        for p in workspace_pipelines:
            self.log.debug(f"- {p}")

        if not workspace_pipelines:
            self.log.debug("- None")

        return workspace_pipelines

    def get_scheduled_pipelines_list(self, client: DatabricksConfig) -> list:
        api_pipelines = self.__get_api_pipeline_list(client)

        pipelines = []
        self.log.debug("Found scheduled pipelines:")
        for p in api_pipelines:
            pipelines.append({p["pipeline_id"]: p["name"]})
            self.log.debug(f"- {p['name']} ({p['pipeline_id']})")

        if not pipelines:
            self.log.debug("- None")

        return pipelines

    def get_obsolete_pipelines_list(
        self, workspace_pipelines: list, scheduled_pipelines: list
    ) -> list:
        obsolete_pipelines = set()

        self.log.debug("Found obsolete pipelines:")
        for p in scheduled_pipelines:
            for p_id, p_name in p.items():
                if p_name not in workspace_pipelines:
                    obsolete_pipelines.add(p_id)
                    self.log.debug(f"- {p_name} ({p_id})")

        if not obsolete_pipelines:
            self.log.debug("- None")

        return obsolete_pipelines

    def delete_obsolete_pipelines(
        self, client: DatabricksConfig, obsolete_pipelines: list
    ) -> None:
        self.log.info("Deleting DLT pipelines for data flows that no longer exist")
        for p in obsolete_pipelines:
            self.__delete_pipeline(client, p)

    def get_new_pipelines_list(
        self, workspace_pipelines: list, scheduled_pipelines: list
    ) -> list:
        for p in scheduled_pipelines:
            for p_id, p_name in p.items():
                if p_name in workspace_pipelines:
                    workspace_pipelines.remove(p_name)

        self.log.debug("Found new pipelines:")
        if not workspace_pipelines:
            self.log.debug("- None")
        else:
            for p in workspace_pipelines:
                self.log.debug("- ", p)

        return workspace_pipelines

    def get_node_types_list(self, client: DatabricksConfig):
        api_node_types = self.__get_api_node_types(client)

        sorted_node_types = sorted(
            api_node_types, key=lambda k: (k["num_cores"], k["memory_mb"])
        )

        node_types = [
            t["node_type_id"]
            for t in sorted_node_types
            if (not t["is_deprecated"]) and ("deprecated" not in t["description"])
        ]
        return node_types

    def __get_cluster_config_for_pipeline(self, pipeline_name: str) -> list:
        for cluster_config in self.config.PipelineConfiguration["clusters"]:
            if matched_config := cluster_config.get(pipeline_name):
                return matched_config

    def create_new_pipelines(
        self, client: DatabricksConfig, new_pipeline_names: list
    ) -> None:
        new_pipelines = []

        self.log.info("Scheduling new pipelines")

        # New pipeline: https://docs.databricks.com/workflows/delta-live-tables/delta-live-tables-api-guide.html#pipeline-spec

        for p in new_pipeline_names:
            cluster_settings = self.__get_cluster_config_for_pipeline(p)

            if not cluster_settings:
                cluster_settings = [{"label": "default", "num_workers": 2}]

            settings = {
                "name": p,
                "storage": None,  # Set per DLT table
                "target": p,
                "clusters": cluster_settings,
                "libraries": [
                    {
                        "notebook": {
                            "path": f"/Repos/{self.user}/{self.repo}/runner/pipeline"
                        }
                    }
                ],
            }

            pipeline_id = self.__create_pipeline(client, settings)
            new_pipelines.append(pipeline_id)

        return new_pipelines


if __name__ == "__main__":
    if app := Application.instance():
        r = Release.get_subapp_instance(app=app)
    else:
        r = Release.instance()

    r.launch_instance()



from .common import *

__all__ = [
    "ExistingSystemPath",
    "ExistingSystemFilePath",
    "ExistingSystemDirectoryPath",
    "ExistingDataPath",
    "ExistingDeltaPath",
    "ExistingTableName",
    "DataPath",
    "TableName",
    "ViewName",
]



import yaml
from traitlets import Bool, Dict, Int, List, Unicode
from traitlets.config.configurable import Configurable


# TODO: Change to custom TraitTypes from validation.common in dictionaries
class PipelineConfiguration(Configurable):
    # Cluster settings https://docs.databricks.com/workflows/delta-live-tables/delta-live-tables-api-guide.html#pipelinesnewcluster
    clusters = List(
        Dict(
            value_trait=List(
                Dict(
                    per_key_traits={
                        "label": Unicode(default_value="default", allow_none=True),
                        "spark_conf": Dict(allow_none=True),
                        "aws_attributes": Dict(allow_none=True),
                        "node_type_id": Unicode(allow_none=True),
                        "driver_node_type_id": Unicode(allow_none=True),
                        "ssh_public_keys": List(
                            trait=Unicode(allow_none=False), allow_none=True
                        ),
                        "custom_tags": Dict(allow_none=True),
                        "cluster_log_conf": Dict(allow_none=True),
                        "spark_env_vars": Dict(allow_none=True),
                        "init_scripts": Dict(allow_none=True),
                        "instance_pool_id": Unicode(allow_none=True),
                        "driver_instance_pool_id": Unicode(allow_none=True),
                        "policy_id": Unicode(allow_none=True),
                        "autoscale": Dict(
                            per_key_traits={
                                "min_workers": Int(allow_none=False),
                                "max_workers": Int(allow_none=False),
                                "mode": Unicode(
                                    default_value="ENHANCED", allow_none=True
                                ),
                            },
                            allow_none=True,
                        ),
                        "num_workers": Int(default_value=2, allow_none=True),
                        "apply_policy_default_values": Bool(allow_none=True),
                    }
                )
            )
        )
    ).tag(config=True)
    sources = List(
        trait=Dict(
            per_key_traits={
                "pipeline_name": Unicode(allow_none=False),
                "from": Unicode(allow_none=False),
                "type": Unicode(allow_none=False),
                "into": Unicode(allow_none=False),
                "params": Unicode(allow_none=True),
            }
        ),
        allow_none=True,
    ).tag(config=True)

    transformations = List(
        trait=Dict(
            per_key_traits={
                "pipeline_name": Unicode(allow_none=False),
                "from": Unicode(allow_none=False),
                "into": Unicode(allow_none=False),
                "column_order": Int(allow_none=True),
                "source_column_name": Unicode(allow_none=True),
                "source_column_type": Unicode(allow_none=True),
                "dest_column_name": Unicode(allow_none=True),
                "dest_column_type": Unicode(allow_none=True),
                "transform_function": Unicode(allow_none=True),
                "sql_query": Unicode(allow_none=True),
                "default_value": Unicode(allow_none=True),
                "not_null": Bool(allow_none=True),
            }
        ),
        allow_none=True,
    ).tag(config=True)

    destinations = List(
        trait=Dict(
            per_key_traits={
                "pipeline_name": Unicode(allow_none=False),
                "from": Unicode(allow_none=False),
                "into": Unicode(allow_none=False),
                "path": Unicode(allow_none=True),
                "keys": List(trait=Unicode(allow_none=False), allow_none=False),
                "partition_by": List(trait=Unicode(allow_none=False), allow_none=True),
                "sequence_by": Unicode(allow_none=False),
            }
        ),
        allow_none=True,
    ).tag(config=True)

    def __str__(self):
        str_repr = {
            "sources": self.sources,
            "transformations": self.transformations,
            "destinations": self.destinations,
        }

        return yaml.dump(str_repr, allow_unicode=True, default_flow_style=False)

    def to_dict(self):
        return {
            "sources": self.sources,
            "transformations": self.transformations,
            "destinations": self.destinations,
        }



import os
import pathlib
import re

from traitlets import Undefined, Unicode

if __package__:
    from delta import DeltaTable
    from pyspark.sql.dataframe import DataFrame

    from pushcart import dbutils, spark


class ExistingSystemPath(Unicode):
    """
    Custom TraitType for storing an existing system path
    """

    def __init__(
        self,
        default_value=Undefined,
        allow_none=False,
        file=False,
        directory=False,
        search_paths=None,
        **kwargs,
    ):
        self.file = file
        self.directory = directory
        self.search_paths = search_paths

        if file and directory:
            raise ValueError("Path cannot be a file and a directory at the same time")

        super(ExistingSystemPath, self).__init__(
            default_value=default_value, allow_none=allow_none, **kwargs
        )

    def validate_system_abs_path(self, path):
        """
        Check whether the path is an existing file, directory, or any kind of path,
        depending on the values of the file and directory attributes. If the path is
        valid, it returns the path, otherwise it returns None
        """
        if (not self.file and not self.directory) and os.path.exists(path):
            return path
        if (self.file and not self.directory) and os.path.isfile(path):
            return path
        if (not self.file and self.directory) and os.path.isdir(path):
            return path

    def validate_system_path(self, value, search_paths=None):
        """
        First normalize path by removing any redundant separators and "." and ".."
        components. If the path is absolute, call the validate_system_abs_path method
        to validate the path. If the path is not absolute, iterate through the search
        paths and append the path to each search path to create an absolute path. Call
        the validate_system_abs_path method to validate each absolute path.
        """
        path = os.path.normpath(value)

        if not search_paths:
            search_paths = [os.getcwd()]

        if os.path.isabs(path):
            return self.validate_system_abs_path(path)
        else:
            for p in search_paths:
                abs_path = os.path.join(os.path.normpath(p), path)

                if found := self.validate_system_abs_path(abs_path):
                    return found

    def validate(self, obj, value):
        value = super(ExistingSystemPath, self).validate(obj, value)

        parsed_value = self.validate_system_path(value, search_paths=self.search_paths)
        if not parsed_value:
            self.error(obj)

        return parsed_value


class ExistingSystemFilePath(ExistingSystemPath):
    """
    Custom TraitType for storing an existing file path on the system
    """

    def __init__(
        self,
        default_value=Undefined,
        allow_none=False,
        search_paths=None,
        **kwargs,
    ):
        super(ExistingSystemFilePath, self).__init__(
            default_value=default_value,
            allow_none=allow_none,
            file=True,
            search_paths=search_paths,
            **kwargs,
        )


class ExistingSystemDirectoryPath(ExistingSystemPath):
    """
    Custom TraitType for storing an existing directory path on the system
    """

    def __init__(
        self,
        default_value=Undefined,
        allow_none=False,
        search_paths=None,
        **kwargs,
    ):
        super(ExistingSystemDirectoryPath, self).__init__(
            default_value=default_value,
            allow_none=allow_none,
            directory=True,
            search_paths=search_paths,
            **kwargs,
        )


class ExistingDataPath(Unicode):
    """
    Custom TraitType for storing an existing path to a file or non-empty directory,
    accessible by Databricks
    """

    default_value = Undefined
    allow_none = False

    def validate_path_exists(self, path):
        path_contents = dbutils.fs.ls(path)

        if path_contents:
            return path

    def validate(self, obj, value):
        value = super(ExistingDataPath, self).validate(obj, value)

        if not self.validate_path_exists(value):
            self.error(obj)

        return value


class ExistingDeltaPath(Unicode):
    """
    Custom TraitType for holding a path to an existing Delta table, accessible from
    Databricks on local storage, DBFS, S3, Azure Storage or Google Storage.
    """

    default_value = Undefined
    allow_none = False

    def validate_path_exists(self, path):
        if DeltaTable.isDeltaTable(spark, path):
            return path

    def validate(self, obj, value):
        value = super(ExistingDeltaPath, self).validate(obj, value)

        if not self.validate_path_exists(value):
            self.error(obj)

        return value


class ExistingTableName(Unicode):
    """
    Custom TraitType for holding the name of an existing Databricks table or view.
    """

    default_value = Undefined
    allow_none = False

    def validate_table_exists(self, table_name):
        if isinstance(spark.table(table_name), DataFrame):
            return table_name

    def validate(self, obj, value):
        value = super(ExistingTableName, self).validate(obj, value)

        if not self.validate_table_exists(value):
            self.error(obj)

        return value


class DataPath(Unicode):
    """
    Custom TraitType for holding a path to data, locally, on DBFS, S3, Azure Storage or
    Google Storage. The path does not necessarily already exist.
    """

    def __init__(self, default_value=Undefined, allow_none=False, **kwargs):
        self.path_regex = [
            re.compile(
                r"^(abfss:\/\/[a-z0-9-]{2,62}@[a-z0-9]{3,24}\.dfs\.core\.windows\.net)\/.*$"
            ),
            re.compile(r"^(s3:\/\/[a-z0-9][a-z0-9.-]{1,61}[a-z0-9])\/.*$"),
            re.compile(r"^(gs://(?:[a-z0-9][a-z0-9-_.]{0,61}[a-z0-9])+)\/.*$"),
            re.compile(r"^(dbfs:)\/.*$"),
            re.compile(r"^(file:)\/.*$"),
        ]
        super(DataPath, self).__init__(
            default_value=default_value, allow_none=allow_none, **kwargs
        )

    def validate_path_syntax(self, path):
        """
        Use a regular expression to check whether the path is in one of file:/,
        dbfs:/, s3:/ or abfss:/ locations. If it is, it split the prefix from the rest
        of the path, then attempt to create a pathlib.Path object from the rest of the
        path. Check whether it is an absolute path and if true, return the path (with
        the prefix added back)
        """
        if not isinstance(path, str):
            return

        prefix = ""
        for regex in self.path_regex:
            if match := regex.match(path):
                prefix, path = match.groups()
                break

        try:
            path = os.path.normpath(path)

            p = pathlib.Path(path)
            if p.is_absolute():
                return prefix + path
        except ValueError:
            pass

    def validate(self, obj, value):
        value = super(DataPath, self).validate(obj, value)

        parsed_value = self.validate_path_syntax(value)
        if not parsed_value:
            self.error(obj)

        return value


class TableName(Unicode):
    """
    Custom TraitType for storing the name of a Spark table, optionally including
    metastore and database. The actual table does not necessarily already exist
    """

    def __init__(self, default_value=Undefined, allow_none=False, **kwargs):
        self.regex_starts_with = re.compile(r"[A-Za-z_]")
        self.regex_allowed_chars = re.compile(r"^[A-Za-z0-9_.]*$")
        self.regex_ends_with = re.compile(r"\.$")

        super(TableName, self).__init__(
            default_value=default_value, allow_none=allow_none, **kwargs
        )

    def validate_table_name(self, value):
        """
        Checks:
        - that the value starts with a letter or underscore
        - that the value only contains letters, digits, underscores, and periods
        - that the value does not end with a period
        - that the length of the value after the last period character is max 64
          characters
        """

        if not isinstance(value, str):
            return

        if not self.regex_starts_with.match(value):
            return

        if not self.regex_allowed_chars.match(value):
            return

        if self.regex_ends_with.match(value):
            return

        if len(value.split(".")[-1]) > 64:
            return

        return value

    def validate(self, obj, value):
        value = super(TableName, self).validate(obj, value)

        if not self.validate_table_name(value):
            self.error(obj)

        return value


class ViewName(Unicode):
    """
    Custom TraitType for storing the name of a Spark view. The actual view does not
    necessarily already exist.
    """

    def __init__(self, default_value=Undefined, allow_none=False, **kwargs):
        self.regex_starts_with = re.compile(r"[A-Za-z_]")
        self.regex_allowed_chars = re.compile(r"^[A-Za-z0-9_]*$")

        super(ViewName, self).__init__(
            default_value=default_value, allow_none=allow_none, **kwargs
        )

    def validate_view_name(self, value):
        """
        Checks:
        - that the value starts with a letter or underscore
        - that the value only contains letters, digits or underscores
        - that the value does not end with a period
        - that the length of the value is max 64 characters
        """

        if not isinstance(value, str):
            return

        if not self.regex_starts_with.match(r"[A-Za-z_]"):
            return

        if not self.regex_allowed_chars.match(r"^[A-Za-z0-9_]*$"):
            return

        if len(value) > 64:
            return

        return value

    def validate(self, obj, value):
        value = super(ViewName, self).validate(obj, value)

        if not self.validate_view_name(value):
            self.error(obj)

        return value



if __package__:
    from pushcart import dbutils

import os
from pathlib import Path

from traitlets import Unicode
from traitlets.config import Application

from pushcart.configuration import PipelineConfiguration


class Pipeline(Application):
    classes = [PipelineConfiguration]
    config_file = Unicode(
        default_value="pipeline_run", help="pipeline configuration file base name"
    ).tag(config=True)

    aliases = {
        "pipeline_name": "Configuration.pipeline_name",
        "sources": "Configuration.sources",
        "transformations": "Configuration.transformations",
        "destinations": "Configuration.destinations",
        ("c", "config-file"): "Pipeline.config_file",
    }

    def initialize(self, argv=None):
        super().initialize(argv=argv)

        if __package__:
            cwd = Path(__file__).parent
        else:
            notebook_path = (
                dbutils.notebook.entry_point.getDbutils()
                .notebook()
                .getContext()
                .notebookPath()
                .get()
            )
            cwd = os.path.join(
                "/Workspace",
                os.path.dirname(os.path.abspath(notebook_path)).lstrip("/"),
            )

        if self.config_file:
            self.load_config_file(
                self.config_file, [os.path.join(cwd, "configuration")]
            )

    def start(self):
        print(Configuration(parent=self))

    @classmethod
    def get_subapp_instance(cls, app: Application) -> Application:
        app.clear_instance()  # since Application is singleton, need to clear main app
        return cls.instance(parent=app)



from functools import lru_cache

try:
    from pyspark.dbutils import DBUtils
except ImportError:
    # TODO: Create DBUtils wrapper using Databricks CLI and keyring
    pass

from pyspark.sql import SparkSession

spark = SparkSession.builder.getOrCreate()

sc = spark.sparkContext
sql = spark.sql

# We'll later overwrite the secrets.get() method of dbutils
dbutils, __dbutils = DBUtils(spark), DBUtils(spark)

__secrets_cache = {}


# TODO: Use keyring to pull secrets from the local keyring when remote
@lru_cache(maxsize=50)
def __get_cached_secret(scope: str, key: str) -> str:
    return __dbutils.secrets.get(scope, key)


dbutils.secrets.get = __get_cached_secret

from .pipeline import *

__all__ = ["Pipeline", "Release", "configuration", "sources"]
